#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for loading water source definitions
"""
import pytest
import yaml
import os

from pyomo.environ import ConcreteModel, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock

from watertap.unit_models.zero_order import FeedZO
from watertap.core import Database, WaterParameterBlock

dbpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
with open(os.path.join(dbpath, "water_sources.yaml"), "r") as f:
    lines = f.read()
    f.close()
source_data = yaml.load(lines, yaml.Loader)

DEFAULT_SOURCE = "seawater"


@pytest.mark.integration
def test_default_source():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(database=m.db)

    m.fs.unit = FeedZO(property_package=m.fs.params)

    for j in m.fs.params.solute_set:
        assert j in source_data[DEFAULT_SOURCE]["solutes"]

    m.fs.unit.load_feed_data_from_database()

    assert pytest.approx(
        source_data[DEFAULT_SOURCE]["default_flow"]["value"], rel=1e-12
    ) == value(m.fs.unit.flow_vol[0])
    assert m.fs.unit.flow_vol[0].fixed

    for j, v in source_data[DEFAULT_SOURCE]["solutes"].items():
        assert pytest.approx(v["value"], rel=1e-12) == value(
            m.fs.unit.conc_mass_comp[0, j]
        )
        assert m.fs.unit.conc_mass_comp[0, j].fixed

    assert_units_consistent(m)


@pytest.mark.integration
@pytest.mark.parametrize(
    "source", list(j for j in source_data.keys() if j != "default")
)
def test_all_sources(source):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(database=m.db, water_source=source)

    m.fs.unit = FeedZO(property_package=m.fs.params)

    for j in m.fs.params.solute_set:
        assert j in source_data[source]["solutes"]

    m.fs.unit.load_feed_data_from_database()

    assert pytest.approx(
        source_data[source]["default_flow"]["value"], rel=1e-12
    ) == value(m.fs.unit.flow_vol[0])
    assert m.fs.unit.flow_vol[0].fixed

    for j, v in source_data[source]["solutes"].items():
        assert pytest.approx(v["value"], rel=1e-12) == value(
            m.fs.unit.conc_mass_comp[0, j]
        )
        assert m.fs.unit.conc_mass_comp[0, j].fixed

        assert j in m.db.component_list.keys()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
    This file demonstrates how to use EDB to create a chemical reactor that
    involves vapor-liquid equilibrium.

    (1) Before we can start, you must install MongoDB (which is installed separately)

        [See more information on the ReadTheDocs under 'Getting Started --> Installing WaterTAP']


    (2) After installing MongoDB, you will need to 'load' the database using the
        command line function 'edb load -b'. This will load the default database
        that WaterTAP is bundled with.

        [NOTE: If you need to 'reload' the database, simply use the command 'edb drop -d electrolytedb'
                in the command line. The database on MongoDB is named "electrolytedb"]

        [NOTE 2: You can invoke the command line utility with the "help" keyword to
                get more information on funtionality. Command: 'edb --help' or 'edb [arg] --help']


    (3) To use EDB in python, start by importing the interface class object 'ElectrolyteDB'


    (4) Invoke the 'ElectrolyteDB' object to connect to the database


    (5) Grab a 'base' for a configuration dictionary, and place it into a class object
        This time, we will grab a base that is for a Liq-Vap problem using FpcTP
        state variables.


    (6) Get the chemcial species/components for a simulation case. There are a number of ways
        to do this. In this example, we will grab them by finding all components that contain
        only specific elements. Then, we add those components and their associated parameters
        to the configuration dictionary being built from the 'base'.

        [NOTE: An alternative method is to provide a list of the names of components you want]


    (7) Get the set of reactions you want in your system and put into a 'base' object.
        In this case, we are getting all reactions associated with a system of water
        and carbonic acid. We should get four reactions now::

            H2O <--> H_+ + OH_-
            H2CO3 <--> H_+ + HCO3_-
            HCO3_- <--> H_+ + CO3_2-
            CO2 + H2O <--> H2CO3


    (8) When using an reactor object in IDAES, you must always provide a 'reaction_config'
        to match with the 'thermo_config'. We can create a base 'reaction' config from
        the database and add reactions to that config in the same way we do for the
        'thermo_config' when adding reactions as inherent.

        [NOTE: If a reaction is added to a 'thermo_config' as 'inherent', it should
               NOT be added to a 'reaction_config' as 'equilibrium']


    (9) Build an equilibrium reactor from the 'thermo_config' and 'reaction_config'
        that were generated from the EDB.

"""

# Import ElectrolyteDB object
from watertap.examples.edb.the_basics import grab_base_reaction_config
from watertap.examples.edb.simple_acid import (
    get_components_and_add_to_idaes_config,
    add_equilibrium_reactions_to_react_base,
    build_equilibrium_model,
)

__author__ = "Austin Ladshaw"


# ========================== (5) ================================
# Grab a new base config for our thermo, but this time we will use
#   one of the newer bases that will use the FpcTP state vars and
#   a Liq-Vap system. This base automatically includes phase equilibrium
#   arguments that IDAES will need to resolve the Liq-Vap problem.
def grab_thermo_Liq_Vap_FpcTP_base(db):
    # Get the base and place into a result object
    base = db.get_base("thermo_Liq_Vap_FpcTP")
    return base


# Run script for testing
def run_vap_liq_with_mockdb(db):
    base_obj = grab_thermo_Liq_Vap_FpcTP_base(db)

    # Our components for this problem are as follows:
    comp_list = ["H2O", "H_+", "OH_-", "H2CO3", "HCO3_-", "CO3_2-", "CO2"]

    base_obj = get_components_and_add_to_idaes_config(db, base_obj, comp_list)

    # Create a reaction config
    react_base = grab_base_reaction_config(db)

    # Add reactions to the reaction base as 'equilibrium'
    react_base = add_equilibrium_reactions_to_react_base(db, react_base, comp_list)

    # Now, we can actually see if we created a correct model by looking
    #       for degrees of freedom, state variables, etc.
    thermo_config = base_obj.idaes_config
    reaction_config = react_base.idaes_config
    model = build_equilibrium_model(thermo_config, reaction_config)

    return model



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import ConcreteModel, value
from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    get_scaling_factor,
    constraint_scaling_transform,
)
from idaes.core.solvers import get_solver
from watertap.examples.flowsheets.full_treatment_train.model_components.eNRTL import (
    entrl_config_FTPx,
    entrl_config_FpcTP,
)
from watertap.examples.flowsheets.full_treatment_train.util import (
    check_scaling,
    solve_block,
)


def simulate_enrtl_FTPx(state_var_args):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = GenericParameterBlock(**entrl_config_FTPx.configuration)

    m.fs.state = m.fs.params.build_state_block(m.fs.time, defined_state=True)

    for (v_name, ind), val in state_var_args.items():
        var = getattr(m.fs.state[0], v_name)
        var[ind].fix(val)
    m.fs.state[0].flow_mol_phase["Liq"].value = 1

    # scale model
    calculate_scaling_factors(m)

    # Regular solve
    solver = get_solver()
    results = solver.solve(m)

    ksp = 3.2e-9  # Gibbs energy gives 3.9e-8, but this fits expectations better
    saturation_index = value(
        m.fs.state[0].act_phase_comp["Liq", "Ca_2+"]
        * m.fs.state[0].act_phase_comp["Liq", "SO4_2-"]
        * m.fs.state[0].act_phase_comp["Liq", "H2O"] ** 2
        / ksp
    )
    return saturation_index


@pytest.mark.component
def test_enrtl_FTPx_0():
    # seawater concentration
    state_var_args = {
        ("temperature", None): 298,
        ("pressure", None): 101325,
        ("flow_mol", None): 100,
        ("mole_frac_comp", "Na_+"): 0.008845,
        ("mole_frac_comp", "Ca_2+"): 0.000174,
        ("mole_frac_comp", "Mg_2+"): 0.001049,
        ("mole_frac_comp", "SO4_2-"): 0.000407,
        ("mole_frac_comp", "Cl_-"): 0.010479,
        ("mole_frac_comp", "H2O"): 0.979046,
    }
    saturation_index = simulate_enrtl_FTPx(state_var_args)
    assert saturation_index == pytest.approx(0.2198, rel=1e-3)


@pytest.mark.component
def test_enrtl_FTPx_1():
    # 2 times seawater concentration
    state_var_args = {
        ("temperature", None): 298,
        ("pressure", None): 101325,
        ("flow_mol", None): 100,
        ("mole_frac_comp", "Na_+"): 0.017327,
        ("mole_frac_comp", "Ca_2+"): 0.000341,
        ("mole_frac_comp", "Mg_2+"): 0.002054,
        ("mole_frac_comp", "SO4_2-"): 0.000796,
        ("mole_frac_comp", "Cl_-"): 0.020529,
        ("mole_frac_comp", "H2O"): 0.958952,
    }
    saturation_index = simulate_enrtl_FTPx(state_var_args)
    assert saturation_index == pytest.approx(0.4333, rel=1e-3)


def simulate_enrtl_FpcTP(state_var_args):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = GenericParameterBlock(**entrl_config_FpcTP.configuration)

    m.fs.state = m.fs.params.build_state_block(m.fs.time, defined_state=True)

    for (v_name, ind), val in state_var_args.items():
        var = getattr(m.fs.state[0], v_name)
        var[ind].fix(val)

    # scale model
    calculate_scaling_factors(m)
    for (ind, c) in m.fs.state[0].true_to_appr_species.items():
        sf = get_scaling_factor(m.fs.state[0].flow_mol_phase_comp_apparent[ind])
        constraint_scaling_transform(c, sf)
    for (ind, c) in m.fs.state[0].appr_mole_frac_constraint.items():
        sf = get_scaling_factor(m.fs.state[0].mole_frac_phase_comp_apparent[ind])
        constraint_scaling_transform(c, sf)

    check_scaling(m)

    solve_block(m)

    ksp = 3.2e-9  # Gibbs energy gives 3.9e-8, but this fits expectations better
    saturation_index = value(
        m.fs.state[0].act_phase_comp["Liq", "Ca_2+"]
        * m.fs.state[0].act_phase_comp["Liq", "SO4_2-"]
        * m.fs.state[0].act_phase_comp["Liq", "H2O"] ** 2
        / ksp
    )
    return saturation_index


@pytest.mark.component
def test_enrtl_FpcTP_1():
    # standard seawater concentration
    feed_flow_mass = 1  # kg/s
    feed_mass_frac_comp = {
        "Na_+": 11122e-6,
        "Ca_2+": 382e-6,
        "Mg_2+": 1394e-6,
        "SO4_2-": 2136e-6,
        "Cl_-": 20316.88e-6,
    }
    feed_mass_frac_comp["H2O"] = 1 - sum(x for x in feed_mass_frac_comp.values())

    mw_comp = {
        "H2O": 18.015e-3,
        "Na_+": 22.990e-3,
        "Ca_2+": 40.078e-3,
        "Mg_2+": 24.305e-3,
        "SO4_2-": 96.06e-3,
        "Cl_-": 35.453e-3,
    }

    feed_flow_mol_comp = {}
    for j in feed_mass_frac_comp:
        feed_flow_mol_comp[j] = feed_flow_mass * feed_mass_frac_comp[j] / mw_comp[j]

    state_var_args = {("temperature", None): 298, ("pressure", None): 101325}
    for j in feed_flow_mol_comp:
        state_var_args[("flow_mol_phase_comp", ("Liq", j))] = feed_flow_mol_comp[j]

    saturation_index = simulate_enrtl_FpcTP(state_var_args)
    assert saturation_index == pytest.approx(0.2200, rel=1e-3)


@pytest.mark.component
def test_enrtl_FpcTP_2():
    # seawater concentration with 50% water removal
    feed_flow_mass = 1  # kg/s
    feed_mass_frac_comp = {
        "Na_+": 11122e-6,
        "Ca_2+": 382e-6,
        "Mg_2+": 1394e-6,
        "SO4_2-": 2136e-6,
        "Cl_-": 20316.88e-6,
    }
    feed_mass_frac_comp["H2O"] = 1 - sum(x for x in feed_mass_frac_comp.values())

    mw_comp = {
        "H2O": 18.015e-3,
        "Na_+": 22.990e-3,
        "Ca_2+": 40.078e-3,
        "Mg_2+": 24.305e-3,
        "SO4_2-": 96.06e-3,
        "Cl_-": 35.453e-3,
    }

    feed_flow_mol_comp = {}
    for j in feed_mass_frac_comp:
        feed_flow_mol_comp[j] = feed_flow_mass * feed_mass_frac_comp[j] / mw_comp[j]
        if j == "H2O":
            feed_flow_mol_comp[j] = feed_flow_mol_comp[j] / 2

    state_var_args = {("temperature", None): 298, ("pressure", None): 101325}
    for j in feed_flow_mol_comp:
        state_var_args[("flow_mol_phase_comp", ("Liq", j))] = feed_flow_mol_comp[j]

    saturation_index = simulate_enrtl_FpcTP(state_var_args)
    assert saturation_index == pytest.approx(0.4344, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""eNRTL property blocks to calculate and constrain gypsum saturation index"""

from pyomo.environ import Block, Var, Expression, Constraint, units as pyunits


def build(m, section="desalination", pretrt_type="NF", **kwargs):
    if section == "desalination":
        m.fs.desal_saturation = Block()
        m.fs.desal_saturation.properties = m.fs.prop_eNRTL.build_state_block([0])
        sb_eNRTL = m.fs.desal_saturation.properties[0]
    elif section == "pretreatment":
        m.fs.pretrt_saturation = Block()
        m.fs.pretrt_saturation.properties = m.fs.prop_eNRTL.build_state_block([0])
        sb_eNRTL = m.fs.pretrt_saturation.properties[0]
    else:
        raise ValueError(
            "{section} is not an expected section for building the saturation index"
            "".format(section=section)
        )

    # populate initial values
    populate_eNRTL_state_vars(sb_eNRTL, base="FpcTP")

    # ksp = 3.9e-9  # Gibbs energy gives 3.9e-8, but this fits expectations better
    ksp = 3.2e-9  # This fits expectations even better

    # constraints
    if section == "desalination":
        sb_dilute = m.fs.tb_pretrt_to_desal.properties_in[0]
        if kwargs["is_twostage"]:
            sb_perm = m.fs.mixer_permeate.mixed_state[0]
            sb_conc = m.fs.RO2.feed_side.properties[0, 1]
            sb_conc_inter = m.fs.RO2.feed_side.properties_interface[0, 1]
        else:
            sb_perm = m.fs.RO.mixed_permeate[0]
            sb_conc = m.fs.RO.feed_side.properties[0, 1]
            sb_conc_inter = m.fs.RO.feed_side.properties_interface[0, 1]

        m.fs.desal_saturation.cp_modulus = Expression(
            expr=sb_conc_inter.conc_mass_phase_comp["Liq", "TDS"]
            / sb_conc.conc_mass_phase_comp["Liq", "TDS"]
        )

        # constraints
        m.fs.desal_saturation.eq_temperature = Constraint(
            expr=sb_eNRTL.temperature == sb_conc.temperature
        )
        m.fs.desal_saturation.eq_pressure = Constraint(
            expr=sb_eNRTL.pressure == sb_conc.pressure
        )

        if pretrt_type == "NF":
            # assumes pretreatment uses the ion property basis
            comp_match_dict = {
                "Na_+": "Na",
                "Ca_2+": "Ca",
                "Mg_2+": "Mg",
                "SO4_2-": "SO4",
                "Cl_-": "Cl",
                "H2O": "H2O",
            }

            @m.fs.desal_saturation.Constraint(comp_match_dict.keys())
            def eq_flow_mol_balance(b, j):
                if j in ["Cl_-", "Na_+"]:
                    bulk_flow = sb_dilute.flow_mol_phase_comp[
                        "Liq", comp_match_dict[j]
                    ] - sb_perm.flow_mass_phase_comp["Liq", "TDS"] / (
                        58.44e-3 * pyunits.kg / pyunits.mol
                    )
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == bulk_flow * m.fs.desal_saturation.cp_modulus
                    )
                elif j in ["Ca_2+", "Mg_2+", "SO4_2-"]:
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == sb_dilute.flow_mol_phase_comp["Liq", comp_match_dict[j]]
                        * m.fs.desal_saturation.cp_modulus
                    )
                elif j == "H2O":
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == sb_conc.flow_mol_phase_comp["Liq", "H2O"]
                    )

        elif pretrt_type == "softening":
            # assumes pretreatment uses the softening property basis
            comp_match_dict = {
                "Na_+": "NaCl",
                "Ca_2+": "Ca(HCO3)2",
                "Mg_2+": "Mg(HCO3)2",
                "SO4_2-": "SO4_2-",
                "Cl_-": "Cl_-",
                "H2O": "H2O",
            }

            @m.fs.desal_saturation.Constraint(comp_match_dict.keys())
            def eq_flow_mol_balance(b, j):
                if j == "Na_+":
                    bulk_flow = sb_dilute.flow_mol_phase_comp[
                        "Liq", "NaCl"
                    ] - sb_perm.flow_mass_phase_comp["Liq", "TDS"] / (
                        58.44e-3 * pyunits.kg / pyunits.mol
                    )
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == bulk_flow * m.fs.desal_saturation.cp_modulus
                    )
                if j in "Cl_-":
                    bulk_flow = (
                        sb_dilute.flow_mol_phase_comp["Liq", "Cl_-"]
                        + sb_dilute.flow_mol_phase_comp["Liq", "NaCl"]
                        - sb_perm.flow_mass_phase_comp["Liq", "TDS"]
                        / (58.44e-3 * pyunits.kg / pyunits.mol)
                    )
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == bulk_flow * m.fs.desal_saturation.cp_modulus
                    )
                elif j in ["Ca_2+", "Mg_2+", "SO4_2-"]:
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == sb_dilute.flow_mol_phase_comp["Liq", comp_match_dict[j]]
                        * m.fs.desal_saturation.cp_modulus
                    )
                elif j == "H2O":
                    return (
                        sb_eNRTL.flow_mol_phase_comp["Liq", j]
                        == sb_conc.flow_mol_phase_comp["Liq", "H2O"]
                    )

        m.fs.desal_saturation.saturation_index = Var(
            initialize=0.5,
            bounds=(0.0, 10),
            units=pyunits.dimensionless,
            doc="Gypsum saturation index",
        )

        m.fs.desal_saturation.eq_saturation_index = Constraint(
            expr=m.fs.desal_saturation.saturation_index
            == sb_eNRTL.act_phase_comp["Liq", "Ca_2+"]
            * sb_eNRTL.act_phase_comp["Liq", "SO4_2-"]
            * sb_eNRTL.act_phase_comp["Liq", "H2O"] ** 2
            / ksp
        )

    elif section == "pretreatment":
        comp_match_dict = {
            "Na_+": "Na",
            "Ca_2+": "Ca",
            "Mg_2+": "Mg",
            "SO4_2-": "SO4",
            "Cl_-": "Cl",
            "H2O": "H2O",
        }

        sb_conc = m.fs.NF.feed_side.properties_out[0]

        m.fs.pretrt_saturation.eq_temperature = Constraint(
            expr=sb_eNRTL.temperature == sb_conc.temperature
        )
        m.fs.pretrt_saturation.eq_pressure = Constraint(
            expr=sb_eNRTL.pressure == sb_conc.pressure
        )

        @m.fs.pretrt_saturation.Constraint(comp_match_dict.keys())
        def eq_flow_mol_balance(b, j):
            return (
                sb_eNRTL.flow_mol_phase_comp["Liq", j]
                == sb_conc.flow_mol_phase_comp["Liq", comp_match_dict[j]]
            )

        m.fs.pretrt_saturation.saturation_index = Var(
            initialize=0.5,
            bounds=(0.0, 1e6),
            units=pyunits.dimensionless,
            doc="Gypsum saturation index",
        )

        m.fs.pretrt_saturation.eq_saturation_index = Constraint(
            expr=m.fs.pretrt_saturation.saturation_index
            == sb_eNRTL.act_phase_comp["Liq", "Ca_2+"]
            * sb_eNRTL.act_phase_comp["Liq", "SO4_2-"]
            * sb_eNRTL.act_phase_comp["Liq", "H2O"] ** 2
            / ksp
        )


def populate_eNRTL_state_vars(blk, base="FpcTP"):
    blk.temperature = 298
    blk.pressure = 101325

    if base == "FpcTP":
        feed_flow_mass = 1  # kg/s
        feed_mass_frac_comp = {
            "Na_+": 11122e-6,
            "Ca_2+": 382e-6,
            "Mg_2+": 1394e-6,
            "SO4_2-": 2136e-6,
            "Cl_-": 20316.88e-6,
        }
        feed_mass_frac_comp["H2O"] = 1 - sum(x for x in feed_mass_frac_comp.values())

        mw_comp = {
            "H2O": 18.015e-3,
            "Na_+": 22.990e-3,
            "Ca_2+": 40.078e-3,
            "Mg_2+": 24.305e-3,
            "SO4_2-": 96.06e-3,
            "Cl_-": 35.453e-3,
        }

        for j in feed_mass_frac_comp:
            blk.flow_mol_phase_comp["Liq", j] = (
                feed_flow_mass * feed_mass_frac_comp[j] / mw_comp[j]
            )
            if j == "H2O":
                blk.flow_mol_phase_comp["Liq", j] /= 2
    elif base == "FTPx":
        blk.mole_frac_comp["Na_+"] = 0.017327
        blk.mole_frac_comp["Ca_2+"] = 0.000341
        blk.mole_frac_comp["Mg_2+"] = 0.002054
        blk.mole_frac_comp["SO4_2-"] = 0.000796
        blk.mole_frac_comp["Cl_-"] = 0.020529
        blk.mole_frac_comp["H2O"] = 0.958952



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os
import idaes.logger as idaeslog

from pyomo.environ import (
    ConcreteModel,
    units as pyunits,
    Expression,
    value,
    TransformationFactory,
)

from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Product
import idaes.core.util.scaling as iscale
from idaes.core import UnitModelCostingBlock

from watertap.core.util.initialization import assert_degrees_of_freedom, check_solve

from watertap.core.wt_database import Database
import watertap.core.zero_order_properties as prop_ZO
from watertap.unit_models.zero_order import (
    FeedZO,
    MagprexZO,
    CentrifugeZO,
    StruviteClassifierZO,
)
from watertap.costing.zero_order_costing import ZeroOrderCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def main():
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m, checkpoint="solve flowsheet after initializing system")
    display_results(m)

    add_costing(m)
    m.fs.costing.initialize()

    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)
    results = solve(m, checkpoint="solve flowsheet with costing")
    display_costing(m)

    return m, results


def build():
    # flowsheet set up
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop = prop_ZO.WaterParameterBlock(solute_list=["phosphates", "struvite"])

    # unit models
    # feed
    m.fs.feed = FeedZO(property_package=m.fs.prop)

    # Magprex reactor treatment for struvite precipitation
    m.fs.magprex = MagprexZO(property_package=m.fs.prop, database=m.db)

    # centrifuge treatment
    m.fs.centrifuge = CentrifugeZO(property_package=m.fs.prop, database=m.db)

    # struvite classifier treatment
    m.fs.classifier = StruviteClassifierZO(property_package=m.fs.prop, database=m.db)

    # product and waste streams
    m.fs.struvite_product = Product(property_package=m.fs.prop)
    m.fs.biosolid_product = Product(property_package=m.fs.prop)
    m.fs.centrate = Product(property_package=m.fs.prop)

    # connections
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.magprex.inlet)
    m.fs.s02 = Arc(source=m.fs.magprex.treated, destination=m.fs.centrifuge.inlet)
    m.fs.s03 = Arc(source=m.fs.centrifuge.treated, destination=m.fs.centrate.inlet)
    m.fs.s04 = Arc(
        source=m.fs.centrifuge.byproduct, destination=m.fs.biosolid_product.inlet
    )
    m.fs.s05 = Arc(source=m.fs.magprex.byproduct, destination=m.fs.classifier.inlet)
    m.fs.s06 = Arc(
        source=m.fs.classifier.outlet, destination=m.fs.struvite_product.inlet
    )

    # expand arcs
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(m):
    # ---specifications---
    # feed
    flow_vol = 126.2 * pyunits.m**3 / pyunits.hr
    conc_mass_phosphates = 350 * pyunits.g / pyunits.m**3
    conc_mass_struvite = 1e-10 * pyunits.g / pyunits.m**3
    m.fs.feed.flow_vol[0].fix(flow_vol)
    m.fs.feed.conc_mass_comp[0, "phosphates"].fix(conc_mass_phosphates)
    m.fs.feed.conc_mass_comp[0, "struvite"].fix(conc_mass_struvite)

    solve(m.fs.feed, checkpoint="solve feed block")

    # Magprex reactor
    m.fs.magprex.load_parameters_from_database(use_default_removal=True)

    # centrifuge
    m.fs.centrifuge.load_parameters_from_database(use_default_removal=True)

    # struvite classifier
    m.fs.classifier.load_parameters_from_database(use_default_removal=True)


def initialize_system(m):
    seq = SequentialDecomposition()
    seq.options.tear_set = []
    seq.options.iterLim = 1
    seq.run(m, lambda u: u.initialize())


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_results(m):
    unit_list = ["feed", "magprex", "centrifuge", "classifier"]
    for u in unit_list:
        m.fs.component(u).report()


def add_costing(m):
    # yaml file path
    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "magprex_case_1575.yaml",
    )

    # add costing block
    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)

    m.fs.magprex.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.centrifuge.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing
    )
    m.fs.classifier.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing
    )

    m.fs.costing.cost_process()

    m.fs.costing.add_electricity_intensity(m.fs.feed.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.feed.properties[0].flow_vol)

    m.fs.costing.annual_water_production = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.centrate.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.annual_water_inlet = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.total_annualized_cost = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
    )

    m.fs.costing.LCOW_with_revenue = Expression(
        expr=(
            (
                m.fs.costing.total_annualized_cost
                + m.fs.costing.aggregate_flow_costs["struvite_product"]
                + m.fs.costing.aggregate_flow_costs["magnesium_chloride"]
                + m.fs.costing.aggregate_flow_costs["polymer"]
            )
            / m.fs.costing.annual_water_production
        ),
        doc="Levelized Cost of Water With Revenue",
    )

    m.fs.costing.LCOT = Expression(
        expr=(m.fs.costing.total_annualized_cost / m.fs.costing.annual_water_inlet),
        doc="Levelized Cost of Treatment Without Revenue",
    )

    m.fs.costing.LCOT_with_revenue = Expression(
        expr=(
            (
                m.fs.costing.total_annualized_cost
                + m.fs.costing.aggregate_flow_costs["struvite_product"]
                + m.fs.costing.aggregate_flow_costs["magnesium_chloride"]
                + m.fs.costing.aggregate_flow_costs["polymer"]
            )
            / m.fs.costing.annual_water_inlet
        ),
        doc="Levelized Cost of Treatment With Revenue",
    )

    m.fs.costing.LCOS = Expression(
        expr=(m.fs.costing.total_annualized_cost)
        / (
            m.fs.costing.utilization_factor
            * pyunits.convert(
                m.fs.struvite_product.properties[0].flow_mass_comp["struvite"],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
        ),
        doc="Levelized cost of struvite",
    )


def display_costing(m):
    print("\n----------Unit Capital Costs----------\n")
    for u in m.fs.costing._registered_unit_costing:
        print(
            u.name,
            " : {price:0.3f} $".format(
                price=value(pyunits.convert(u.capital_cost, to_units=pyunits.USD_2018))
            ),
        )
    print("\n----------Utility Costs----------\n")
    for f in m.fs.costing.used_flows:
        print(
            f,
            " :    {price:0.3f} $/year".format(
                price=value(
                    pyunits.convert(
                        m.fs.costing.aggregate_flow_costs[f],
                        to_units=pyunits.USD_2018 / pyunits.year,
                    )
                )
            ),
        )

    magnesium_chloride_cost = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["magnesium_chloride"],
            to_units=pyunits.USD_2018 / pyunits.year,
        )
        / pyunits.convert(
            m.fs.centrate.properties[0].flow_vol, to_units=pyunits.m**3 / pyunits.year
        )
    )
    print(f"magnesium chloride: { magnesium_chloride_cost: .4f} $/m3 of centrate")

    polymer_cost = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["polymer"],
            to_units=pyunits.USD_2018 / pyunits.year,
        )
        / pyunits.convert(
            m.fs.centrate.properties[0].flow_vol, to_units=pyunits.m**3 / pyunits.year
        )
    )
    print(f"polymer: { polymer_cost: .4f} $/m3 of centrate")

    struvite_cost = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["struvite_product"],
            to_units=pyunits.USD_2018 / pyunits.year,
        )
        / pyunits.convert(
            m.fs.centrate.properties[0].flow_vol, to_units=pyunits.m**3 / pyunits.year
        )
    )
    print(f"struvite: { struvite_cost: .4f} $/m3 of centrate")

    print("\n----------Total Costs----------\n")

    total_capital_cost = value(
        pyunits.convert(m.fs.costing.total_capital_cost, to_units=pyunits.USD_2018)
    )
    print(f"Total Capital Costs: {total_capital_cost:.3f} $")

    total_operating_cost = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.USD_2018 / pyunits.year
        )
    )
    print(f"Total Operating Costs: {total_operating_cost:.3f} $/year")
    electricity_intensity = value(
        pyunits.convert(
            m.fs.costing.electricity_intensity, to_units=pyunits.kWh / pyunits.m**3
        )
    )

    print(f"Electricity Intensity: {electricity_intensity:.3f} kWh/m^3")

    print("\n----------Levelized Costs----------\n")

    LCOW_with_revenue = value(
        pyunits.convert(
            m.fs.costing.LCOW_with_revenue,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(
        f"Levelized Cost of Water with Revenue: {LCOW_with_revenue:.4f} $/m3 of centrate"
    )
    LCOW = value(
        pyunits.convert(m.fs.costing.LCOW, to_units=pyunits.USD_2018 / pyunits.m**3)
    )
    print(f"Levelized Cost of Water Without Revenue: {LCOW:.3f} $/m^3 of centrate")

    LCOT = value(
        pyunits.convert(m.fs.costing.LCOT, to_units=pyunits.USD_2020 / pyunits.m**3)
    )
    print(f"Levelized Cost of Treatment: {LCOT:.3f} $/m^3")
    LCOT_with_revenue = value(
        pyunits.convert(
            m.fs.costing.LCOT_with_revenue, to_units=pyunits.USD_2020 / pyunits.m**3
        )
    )
    print(f"Levelized Cost of Treatment With Revenue: {LCOT_with_revenue:.3f} $/m^3")

    LCOS = value(
        pyunits.convert(m.fs.costing.LCOS, to_units=pyunits.USD_2018 / pyunits.kg)
    )
    print(f"Levelized Cost of Struvite: {LCOS:.3f} $/kg")


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
    Scaling utilities for flowsheets involving electrolyte chemistry

    NOTE: These are specific for the demonstration flowsheet examples
"""

from idaes.core.util import scaling as iscale
from idaes.core.util.initialization import fix_state_vars, revert_state_vars

# Import specific pyomo objects
from pyomo.environ import value

__author__ = "Austin Ladshaw"

# Given a unit model with corresponding set of reaction params and reaction config dictionary
#   find an approximate set of state_args and/or stoich_extents to be used in initialization
#   and scaling for electrolyte systems.
#
#       Assumes using FTPx as the state_vars
def approximate_chemical_state_args(
    unit, rxn_params, reaction_config, contains_stoich_reactions=False
):
    state_args = {}
    stoich_extents = {}

    # Set bulk values to their inlets
    state_args["pressure"] = unit.inlet.pressure[0].value
    state_args["temperature"] = unit.inlet.temperature[0].value
    state_args["flow_mol"] = unit.inlet.flow_mol[0].value

    # Set species based on inlets (and outlets for stoich reaction)
    state_args["mole_frac_comp"] = {}
    min = 1e-08
    for i in unit.control_volume.properties_in[0.0].mole_frac_comp:
        # Set state args to inlets on first pass
        if unit.inlet.mole_frac_comp[0, i].value > min:
            state_args["mole_frac_comp"][i] = unit.inlet.mole_frac_comp[0, i].value
        else:
            state_args["mole_frac_comp"][i] = min

    # Iterate through outlet mole fractions and note the fixed variables
    fixed = {}
    for i, species in unit.outlet.mole_frac_comp:
        if unit.outlet.mole_frac_comp[i, species].is_fixed():
            fixed[species] = True
            state_args["mole_frac_comp"][species] = unit.outlet.mole_frac_comp[
                0, species
            ].value

    return state_args, stoich_extents


# Perform scaling transformations for equilibrium reactions (if they exist)
def calculate_chemical_scaling_factors_for_equilibrium_log_reactions(unit, rxn_params):
    try:
        for i in unit.control_volume.equilibrium_reaction_extent_index:
            if i[1] != "dummy":
                scale = value(unit.control_volume.reactions[0.0].k_eq[i[1]].expr)
                iscale.set_scaling_factor(
                    unit.control_volume.equilibrium_reaction_extent[0.0, i[1]],
                    10 / scale,
                )
                iscale.constraint_scaling_transform(
                    unit.control_volume.reactions[0.0].equilibrium_constraint[i[1]], 0.1
                )
    except:
        pass


# # Perform scaling transformations for mass balances
def calculate_chemical_scaling_factors_for_material_balances(unit):
    # Next, try adding scaling for species
    min = 1e-3
    for i in unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        # i[0] = phase, i[1] = species
        if unit.inlet.mole_frac_comp[0, i[1]].value > min:
            scale = unit.inlet.mole_frac_comp[0, i[1]].value
        else:
            scale = min
        iscale.set_scaling_factor(
            unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10 / scale
        )
        iscale.set_scaling_factor(
            unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10 / scale
        )
        iscale.set_scaling_factor(
            unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10 / scale
        )
        iscale.constraint_scaling_transform(
            unit.control_volume.properties_out[0.0].component_flow_balances[i[1]],
            10 / scale,
        )
        iscale.constraint_scaling_transform(
            unit.control_volume.material_balances[0.0, i[1]], 10 / scale
        )


# # Perform scaling transformations for energy balances
def calculate_chemical_scaling_factors_for_energy_balances(unit):
    max = 1
    min = 1
    try:
        for phase in unit.control_volume.properties_in[0.0].enth_mol_phase:
            val = abs(
                value(unit.control_volume.properties_in[0.0].enth_mol_phase[phase].expr)
            )
            if val >= max:
                max = val
            if val <= min:
                val = min
            iscale.set_scaling_factor(
                unit.control_volume.properties_in[0.0]._enthalpy_flow_term[phase],
                10 / val,
            )
            iscale.set_scaling_factor(
                unit.control_volume.properties_out[0.0]._enthalpy_flow_term[phase],
                10 / val,
            )

        iscale.constraint_scaling_transform(
            unit.control_volume.enthalpy_balances[0.0], 10 / max
        )
    except:
        pass


# Serially calculate all scaling factors needed
def calculate_chemical_scaling_factors(
    unit, thermo_params, rxn_params, state_args, output_jac=False
):
    calculate_chemical_scaling_factors_for_equilibrium_log_reactions(unit, rxn_params)
    calculate_chemical_scaling_factors_for_energy_balances(unit)
    calculate_chemical_scaling_factors_for_material_balances(unit)

    flags = fix_state_vars(unit.control_volume.properties_out, state_args)
    revert_state_vars(unit.control_volume.properties_out, flags)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import value
from watertap.examples.flowsheets.full_treatment_train.analysis import (
    flowsheet_NF,
    flowsheet_NF_no_bypass,
    flowsheet_single_stage,
    flowsheet_two_stage,
    flowsheet_NF_two_stage,
    flowsheet_softening,
    flowsheet_softening_two_stage,
)


@pytest.mark.component
def test_flowsheet_NF():
    m = flowsheet_NF.optimize_flowsheet(system_recovery=0.5)
    assert value(
        m.fs.tb_pretrt_to_desal.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(0.7675, rel=1e-3)
    assert value(
        m.fs.tb_pretrt_to_desal.properties_in[0].flow_mass_phase_comp["Liq", "Ca"]
    ) == pytest.approx(1.175e-4, rel=1e-3)


@pytest.mark.component
def test_flowsheet_NF_no_bypass():
    m = flowsheet_NF_no_bypass.solve_flowsheet()
    assert value(
        m.fs.tb_pretrt_to_desal.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(0.29225, rel=1e-3)
    assert value(
        m.fs.tb_pretrt_to_desal.properties_in[0].flow_mass_phase_comp["Liq", "Ca"]
    ) == pytest.approx(2.413e-5, rel=1e-3)
    m.fs.NF.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(0.5)
    flowsheet_NF_no_bypass.simulate(m, **{"unfix_nf_area": True})
    assert value(
        m.fs.tb_pretrt_to_desal.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(0.4823, rel=1e-3)


@pytest.mark.component
def test_flowsheet_single_stage():
    desal_kwargs = flowsheet_single_stage.desal_kwargs
    m = flowsheet_single_stage.optimize_flowsheet(system_recovery=0.5, **desal_kwargs)
    assert value(m.fs.costing.LCOW) == pytest.approx(0.4996, rel=1e-3)


@pytest.mark.component
def test_flowsheet_two_stage():
    desal_kwargs = flowsheet_two_stage.desal_kwargs
    m = flowsheet_two_stage.optimize_flowsheet(system_recovery=0.65, **desal_kwargs)
    assert value(m.fs.costing.LCOW) == pytest.approx(0.5435, rel=1e-3)


@pytest.mark.component
def test_flowsheet_NF_two_stage():
    desal_kwargs = flowsheet_two_stage.desal_kwargs
    m = flowsheet_NF_two_stage.optimize_flowsheet(system_recovery=0.70, **desal_kwargs)
    assert value(m.fs.costing.LCOW) == pytest.approx(0.5502, rel=1e-3)


@pytest.mark.component
def test_flowsheet_softening():
    m = flowsheet_softening.solve_flowsheet()
    assert value(m.fs.costing.LCOW) == pytest.approx(0.3837, rel=1e-3)


@pytest.mark.component
def test_flowsheet_softening_two_stage():
    desal_kwargs = flowsheet_two_stage.desal_kwargs
    m = flowsheet_softening_two_stage.optimize_flowsheet(
        system_recovery=0.80, **desal_kwargs
    )
    assert value(m.fs.costing.LCOW) == pytest.approx(0.92855, rel=1e-3)


@pytest.mark.component
def test_flowsheet_1DRO():
    desal_kwargs = flowsheet_two_stage.desal_kwargs
    desal_kwargs["RO_type"] = "1D"
    m = flowsheet_two_stage.optimize_flowsheet(system_recovery=0.70, **desal_kwargs)
    assert value(m.fs.costing.LCOW) == pytest.approx(0.5963, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from .control_volume_isothermal import ControlVolume0DBlock, ControlVolume1DBlock
from .initialization_mixin import InitializationMixin
from .membrane_channel_base import (
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
    FrictionFactor,
)
from .membrane_channel0d import MembraneChannel0DBlock
from .membrane_channel1d import MembraneChannel1DBlock
from .wt_database import Database
from .zero_order_base import ZeroOrderBaseData
from .zero_order_properties import WaterParameterBlock, WaterStateBlock
from .zero_order_electricity import constant_intensity, pump_electricity
from .zero_order_pt import build_pt
from .zero_order_sido import build_sido
from .zero_order_sido_reactive import build_sido_reactive
from .zero_order_siso import build_siso
from .zero_order_diso import build_diso



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import os
import numpy as np
import pyomo.environ as pyo
import yaml

from pyomo.environ import value

from watertap.tools.parameter_sweep.parameter_sweep_reader import (
    ParameterSweepReader,
    get_sweep_params_from_yaml,
    set_defaults_from_yaml,
)


class TestInputParser:
    @pytest.fixture(scope="class")
    def model(self):
        m = pyo.ConcreteModel()
        m.fs = pyo.Block()
        m.fs.x = pyo.Var(initialize=-0.5)
        m.fs.a = pyo.Param(initialize=1.0, mutable=True)
        m.fs.b = pyo.Param(initialize=2.0, mutable=False)
        m.fs.sum = pyo.Expression(expr=m.fs.a + m.fs.x)
        m.fs.lim = pyo.Constraint(expr=m.fs.sum >= 0.0)
        m.fs.obj = pyo.Objective(expr=m.fs.x, sense=pyo.minimize)

        return m

    @pytest.fixture(scope="class")
    def get_simple_yaml_file(self):
        filename = "temp.yaml"
        dict_to_write = {"a_val": 1.1, "b_val": 2.2, "c_val": 3.3, "d_val": 4.4}

        with open(filename, "w") as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write

    @pytest.fixture(scope="class")
    def get_param_yaml_file(self):
        filename = "temp.yaml"
        dict_to_write = {
            "a_val": {
                "type": "LinearSample",
                "param": "fs.a",
                "lower_limit": 10,
                "upper_limit": 20,
                "num_samples": 5,
            },
            "b_val": {
                "type": "UniformSample",
                "param": "fs.a",
                "lower_limit": 10,
                "upper_limit": 20,
                "num_samples": 5,
            },
            "c_val": {
                "type": "NormalSample",
                "param": "fs.a",
                "mean": 10,
                "std": 2,
                "num_samples": 5,
            },
            "d_val": {
                "type": "LatinHypercubeSample",
                "param": "fs.a",
                "lower_limit": 10,
                "upper_limit": 20,
                "num_samples": 5,
            },
            "e_val": {
                "type": "GeomSample",
                "param": "fs.a",
                "lower_limit": 1,
                "upper_limit": 10,
                "num_samples": 3,
            },
            "f_val": {
                "type": "ReverseGeomSample",
                "param": "fs.a",
                "lower_limit": 1,
                "upper_limit": 10,
                "num_samples": 3,
            },
            "g_ val": {
                "type": "PredeterminedFixedSample",
                "param": "fs.a",
                "array": [1, 2, 3, 4, 5],
            },
            "h_val": {
                "type": "PredeterminedRandomSample",
                "param": "fs.a",
                "array": [1, 2, 3, 4, 5],
            },
        }

        with open(filename, "w") as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write

    @pytest.fixture(scope="class")
    def get_param_yaml_file_error(self):
        filename = "temp.yaml"
        dict_to_write = {"a_val": {"param": "nonexistent_variable"}}

        with open(filename, "w") as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write

    @pytest.fixture(scope="class")
    def get_default_yaml_file(self, model):
        filename = "temp.yaml"

        m = model

        dict_to_write = {}

        for pyo_obj in m.component_data_objects((pyo.Var, pyo.Param), active=True):
            for k in range(10):
                dict_to_write[pyo_obj.name] = np.random.rand()

        with open(filename, "w") as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write

    @pytest.fixture(scope="class")
    def get_default_yaml_file_error(self, model):
        filename = "temp.yaml"

        dict_to_write = {"nonexistent_variable": 1.0}

        with open(filename, "w") as fp:
            yaml.dump(dict_to_write, fp, sort_keys=False)

        return filename, dict_to_write

    @pytest.mark.unit
    def test_yaml_to_dict(self, get_simple_yaml_file):
        filename, truth_values = get_simple_yaml_file

        psr = ParameterSweepReader()
        values = psr._yaml_to_dict(filename)

        for key, truth_value in truth_values.items():
            assert key in values
            assert values[key] == truth_value

    @pytest.mark.unit
    def test_yaml_to_dict_error(self):
        psr = ParameterSweepReader()
        with pytest.raises(Exception):
            values = psr._yaml_to_dict("nonexistent_file.yaml")

    @pytest.mark.unit
    def test_get_sweep_params_from_yaml(self, model, get_param_yaml_file):
        m = model
        filename, truth_values = get_param_yaml_file

        values = get_sweep_params_from_yaml(m, filename)

        for key, truth_value in truth_values.items():
            assert key in values

    @pytest.mark.unit
    def test_get_sweep_params_from_yaml_error(self, model, get_param_yaml_file_error):
        m = model
        filename, truth_values = get_param_yaml_file_error

        psr = ParameterSweepReader()
        with pytest.raises(ValueError):
            psr.get_sweep_params_from_yaml(m, filename)

    @pytest.mark.unit
    def test_set_defaults_from_yaml(self, model, get_default_yaml_file):
        m = model
        filename, truth_values = get_default_yaml_file

        set_defaults_from_yaml(m, filename)

        for key, truth_value in truth_values.items():
            component = m.find_component(key)
            if component is m.fs.b:  # don't set non-mutable params
                assert value(component) == 2.0
            else:
                assert value(component) == truth_value

    @pytest.mark.unit
    def test_set_defaults_from_yaml_error(self, model, get_default_yaml_file_error):
        m = model
        filename, truth_values = get_default_yaml_file_error

        psr = ParameterSweepReader()
        with pytest.raises(ValueError):
            psr.set_defaults_from_yaml(m, filename)

        os.remove(filename)

    @pytest.mark.unit
    def test_dict_to_diff(self, model):
        diff_spec_dict = {
            "fs.x": {
                "diff_mode": "percentile",
                "diff_sample_type": "UniformSample",
                "param": "fs.x",
                "relative_lb": -0.01,
                "relative_ub": -0.01,
                "nominal_lb": 15,
                "nominal_ub": 30,
                "num_samples": 1,
            },
        }
        diff_spec = ParameterSweepReader()._dict_to_diff_spec(model, diff_spec_dict)
        for key in diff_spec_dict:
            assert key in diff_spec
        diff_spec_dict = {
            "fs.x": {
                "diff_mode": "percentile",
                "diff_sample_type": "LinearSample",
                "param": "fs.x",
                "relative_lb": -0.01,
                "relative_ub": -0.01,
                "nominal_lb": 15,
                "nominal_ub": 30,
                "num_samples": 1,
            },
        }
        diff_spec = ParameterSweepReader()._dict_to_diff_spec(model, diff_spec_dict)
        for key in diff_spec_dict:
            assert key in diff_spec
        diff_spec_dict = {
            "fs.x": {
                "diff_mode": "sum",
                "diff_sample_type": "UniformSample",
                "param": "fs.x",
                "relative_lb": -0.01,
                "relative_ub": -0.01,
                "nominal_lb": 15,
                "nominal_ub": 30,
                "num_samples": 1,
            },
        }
        diff_spec = ParameterSweepReader()._dict_to_diff_spec(model, diff_spec_dict)
        diff_spec_dict["fs.x"]["relative_lb"] = None
        diff_spec_dict["fs.x"]["relative_ub"] = None
        for key in diff_spec_dict:
            assert key in diff_spec
        diff_spec_dict = {
            "fs.x": {
                "diff_mode": "sum",
                "diff_sample_type": "LinearSample",
                "param": "fs.x",
                "relative_lb": -0.01,
                "relative_ub": -0.01,
                "nominal_lb": 15,
                "nominal_ub": 30,
                "num_samples": 1,
            },
        }
        diff_spec = ParameterSweepReader()._dict_to_diff_spec(model, diff_spec_dict)
        diff_spec_dict["fs.x"]["relative_lb"] = None
        diff_spec_dict["fs.x"]["relative_ub"] = None
        for key in diff_spec_dict:
            assert key in diff_spec
        diff_spec_dict = {
            "fs.x": {
                "diff_mode": "percentile",
                "diff_sample_type": "NormalSample",
                "param": "fs.x",
                "std_dev": 0.4,
            },
        }
        diff_spec = ParameterSweepReader()._dict_to_diff_spec(model, diff_spec_dict)
        for key in diff_spec_dict:
            assert key in diff_spec



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order membrane bioreactor model
"""
import pytest


from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import MBRZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestMBRZOdefault:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "tss",
                "nonvolatile_toc",
                "toc",
                "eeq",
                "viruses_enteric",
                "total_coliforms_fecal_ecoli",
                "cryptosporidium",
            ]
        )

        m.fs.unit = MBRZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.elec_coeff_1, Var)
        assert isinstance(model.fs.unit.elec_coeff_2, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_intensity, Var)
        assert isinstance(model.fs.unit.electricity_intensity_constraint, Constraint)
        assert isinstance(model.fs.unit.electricity_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("mbr")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.elec_coeff_1.fixed
        assert model.fs.unit.elec_coeff_1.value == data["elec_coeff_1"]["value"]

        assert model.fs.unit.elec_coeff_2.fixed
        assert model.fs.unit.elec_coeff_2.value == data["elec_coeff_2"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(1.007, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["total_coliforms_fecal_ecoli"]
        )
        assert pytest.approx(0.99305, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["cryptosporidium"]
        )
        assert pytest.approx(1.001222762, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.49939, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.39951, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(0.29110405, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(0.11985345, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(0.0099878, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(0.001200532, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"
            ]
        )
        assert pytest.approx(9.9878e-05, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["cryptosporidium"]
        )
        assert pytest.approx(0.0057772, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(86.547, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(103.8558564, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(122.6433808, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(152.3219227, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(171.36216303, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(172.8850361, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"
            ]
        )
        assert pytest.approx(173.0757847, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["cryptosporidium"]
        )
        assert pytest.approx(2946.15766, abs=1e-5) == value(
            model.fs.unit.electricity[0]
        )
        assert pytest.approx(0.812688, abs=1e-5) == value(
            model.fs.unit.electricity_intensity[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


class TestMBRZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "tss",
                "nonvolatile_toc",
                "toc",
                "eeq",
                "viruses_enteric",
                "total_coliforms_fecal_ecoli",
                "cryptosporidium",
                "foo",
            ]
        )

        m.fs.unit = MBRZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.elec_coeff_1, Var)
        assert isinstance(model.fs.unit.elec_coeff_2, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_intensity, Var)
        assert isinstance(model.fs.unit.electricity_intensity_constraint, Constraint)
        assert isinstance(model.fs.unit.electricity_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("mbr")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.elec_coeff_1.fixed
        assert model.fs.unit.elec_coeff_1.value == data["elec_coeff_1"]["value"]

        assert model.fs.unit.elec_coeff_2.fixed
        assert model.fs.unit.elec_coeff_2.value == data["elec_coeff_2"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(1.008, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["total_coliforms_fecal_ecoli"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["cryptosporidium"]
        )
        assert pytest.approx(0.99206, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(1.002222762018698, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.49889, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.39911, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(0.290813, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(0.119734, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(0.00997782, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(0.00119934, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"
            ]
        )
        assert pytest.approx(9.9788e-05, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["cryptosporidium"]
        )
        assert pytest.approx(0.99778, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(0.0057772, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(86.547, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(103.8558564, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]
        )
        assert pytest.approx(122.6433808, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(152.3219227, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(171.36216303, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["viruses_enteric"]
        )
        assert pytest.approx(172.8850361, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"
            ]
        )
        assert pytest.approx(173.0757847, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["cryptosporidium"]
        )
        assert pytest.approx(1.73093e-6, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(2948.205338, abs=1e-5) == value(
            model.fs.unit.electricity[0]
        )
        assert pytest.approx(0.812446, abs=1e-5) == value(
            model.fs.unit.electricity_intensity[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


db = Database()
params = db._get_technology("mbr")


class TestMBRZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "tss",
                "nonvolatile_toc",
                "toc",
                "eeq",
                "viruses_enteric",
                "total_coliforms_fecal_ecoli",
                "cryptosporidium",
            ]
        )

        m.fs.unit = MBRZO(property_package=m.fs.params, database=db)

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("mbr", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.elec_coeff_1.fixed
        assert model.fs.unit.elec_coeff_1.value == data["elec_coeff_1"]["value"]

        assert model.fs.unit.elec_coeff_2.fixed
        assert model.fs.unit.elec_coeff_2.value == data["elec_coeff_2"]["value"]


@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tds"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = MBRZO(
        property_package=m.fs.params, database=m.db, process_subtype=subtype
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tds"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.mbr, Block)
    assert isinstance(m.fs.costing.mbr.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.mbr.capital_b_parameter, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
__author__ = "Chenyu Wang"

import pyomo.environ as pyo
from pyomo.environ import (
    units,
    value,
    assert_optimal_termination,
    units as pyunits,
)
from pyomo.network import Arc
from idaes.core import (
    FlowsheetBlock,
    UnitModelCostingBlock,
)
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from watertap.unit_models.anaerobic_digestor import AD
from watertap.property_models.anaerobic_digestion.modified_adm1_properties import (
    ModifiedADM1ParameterBlock,
)
from watertap.property_models.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)
from watertap.property_models.anaerobic_digestion.modified_adm1_reactions import (
    ModifiedADM1ReactionParameterBlock,
)
from watertap.property_models.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from watertap.unit_models.translators.translator_adm1_asm2d import (
    Translator_ADM1_ASM2D,
)
from watertap.unit_models.electroNP_ZO import ElectroNPZO
from idaes.core.util.tables import (
    create_stream_table_dataframe,
    stream_table_dataframe_to_string,
)
from idaes.core.util.initialization import propagate_state
from watertap.core.util.initialization import check_solve
from watertap.costing import WaterTAPCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_flowsheet():
    # flowsheet set up
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props_ADM1 = ModifiedADM1ParameterBlock()
    m.fs.props_vap_ADM1 = ADM1_vaporParameterBlock()
    m.fs.rxn_props_ADM1 = ModifiedADM1ReactionParameterBlock(
        property_package=m.fs.props_ADM1
    )
    m.fs.props_ASM2D = ModifiedASM2dParameterBlock()
    m.fs.costing = WaterTAPCosting()

    # Unit models
    m.fs.AD = AD(
        liquid_property_package=m.fs.props_ADM1,
        vapor_property_package=m.fs.props_vap_ADM1,
        reaction_package=m.fs.rxn_props_ADM1,
        has_heat_transfer=True,
        has_pressure_change=False,
    )
    m.fs.AD.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    m.fs.translator_adm1_asm2d = Translator_ADM1_ASM2D(
        inlet_property_package=m.fs.props_ADM1,
        outlet_property_package=m.fs.props_ASM2D,
        reaction_package=m.fs.rxn_props_ADM1,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    m.fs.electroNP = ElectroNPZO(property_package=m.fs.props_ASM2D)
    m.fs.electroNP.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.electroNP.treated.flow_vol[0])
    m.fs.costing.add_LCOW(m.fs.AD.inlet.flow_vol[0])

    # connections
    m.fs.stream_adm1_translator = Arc(
        source=m.fs.AD.liquid_outlet, destination=m.fs.translator_adm1_asm2d.inlet
    )
    m.fs.stream_translator_electroNP = Arc(
        source=m.fs.translator_adm1_asm2d.outlet, destination=m.fs.electroNP.inlet
    )
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Feed conditions based on mass balance in Flores-Alsina, where 0 terms are expressed as 1e-9
    m.fs.AD.inlet.flow_vol[0].fix(
        170 * units.m**3 / units.day
    )  # Double check this value
    m.fs.AD.inlet.temperature.fix(308.15)
    m.fs.AD.inlet.pressure.fix(101325)

    m.fs.AD.inlet.conc_mass_comp[0, "S_su"].fix(0.034597)
    m.fs.AD.inlet.conc_mass_comp[0, "S_aa"].fix(0.015037)
    m.fs.AD.inlet.conc_mass_comp[0, "S_fa"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "S_va"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "S_bu"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "S_pro"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "S_ac"].fix(0.025072)
    m.fs.AD.inlet.conc_mass_comp[0, "S_h2"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "S_ch4"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "S_IC"].fix(0.34628)
    m.fs.AD.inlet.conc_mass_comp[0, "S_IN"].fix(0.60014)
    m.fs.AD.inlet.conc_mass_comp[0, "S_IP"].fix(0.22677)
    m.fs.AD.inlet.conc_mass_comp[0, "S_I"].fix(0.026599)

    m.fs.AD.inlet.conc_mass_comp[0, "X_ch"].fix(7.3687)
    m.fs.AD.inlet.conc_mass_comp[0, "X_pr"].fix(7.7308)
    m.fs.AD.inlet.conc_mass_comp[0, "X_li"].fix(10.3288)
    m.fs.AD.inlet.conc_mass_comp[0, "X_su"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_aa"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_fa"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_c4"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_pro"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_ac"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_h2"].fix(1e-6)
    m.fs.AD.inlet.conc_mass_comp[0, "X_I"].fix(12.7727)
    m.fs.AD.inlet.conc_mass_comp[0, "X_PHA"].fix(0.0022493)
    m.fs.AD.inlet.conc_mass_comp[0, "X_PP"].fix(1.04110)
    m.fs.AD.inlet.conc_mass_comp[0, "X_PAO"].fix(3.4655)
    m.fs.AD.inlet.conc_mass_comp[0, "S_K"].fix(0.02268)
    m.fs.AD.inlet.conc_mass_comp[0, "S_Mg"].fix(0.02893)

    m.fs.AD.inlet.cations[0].fix(0.04)
    m.fs.AD.inlet.anions[0].fix(0.02)

    m.fs.AD.volume_liquid.fix(3400)
    m.fs.AD.volume_vapor.fix(300)
    m.fs.AD.liquid_outlet.temperature.fix(308.15)

    # ElectroNP
    m.fs.electroNP.energy_electric_flow_mass.fix(0.044 * units.kWh / units.kg)
    m.fs.electroNP.magnesium_chloride_dosage.fix(0.388)

    # Costing
    m.fs.costing.electroNP.phosphorus_recovery_value = 0

    # scaling
    for var in m.fs.component_data_objects(pyo.Var, descend_into=True):
        if "flow_vol" in var.name:
            iscale.set_scaling_factor(var, 1e3)
        if "temperature" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "pressure" in var.name:
            iscale.set_scaling_factor(var, 1e-3)
        if "alkalinity" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "conc_mass_comp" in var.name:
            iscale.set_scaling_factor(var, 1e1)
        if "conc_mass_comp[S_IN]" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "conc_mass_comp[S_IP]" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "conc_mass_comp[S_PO4]" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "conc_mass_comp[S_NH4]" in var.name:
            iscale.set_scaling_factor(var, 1e1)
        if "conc_mass_comp[S_F]" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "conc_mass_comp[X_I]" in var.name:
            iscale.set_scaling_factor(var, 1e-1)

    iscale.calculate_scaling_factors(m)

    iscale.set_scaling_factor(m.fs.electroNP.byproduct.flow_vol[0.0], 1e7)
    iscale.set_scaling_factor(m.fs.AD.vapor_phase[0].pressure_sat, 1e-3)

    m.fs.AD.initialize(outlvl=idaeslog.INFO_HIGH)
    propagate_state(m.fs.stream_adm1_translator)
    m.fs.translator_adm1_asm2d.initialize(outlvl=idaeslog.INFO_HIGH)
    propagate_state(m.fs.stream_translator_electroNP)
    m.fs.electroNP.initialize(outlvl=idaeslog.INFO_HIGH)
    m.fs.costing.initialize()

    results = solve(m, tee=True)
    return m, results


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_costing(m):
    print("\n----------Capital Cost----------")
    total_capital_cost = value(
        pyunits.convert(m.fs.costing.total_capital_cost, to_units=pyunits.USD_2018)
    )
    normalized_capex = total_capital_cost / value(
        pyunits.convert(m.fs.AD.inlet.flow_vol[0], to_units=pyunits.m**3 / pyunits.hr)
    )
    print(f"Total Capital Costs: {total_capital_cost:.3f} $")
    print(f"Normalized Capital Costs: {normalized_capex:.3f} $/m3/hr")
    print("Capital Cost Breakdown")
    for u in m.fs.costing._registered_unit_costing:
        print(
            u.name,
            " : {price:0.3f} $".format(
                price=value(pyunits.convert(u.capital_cost, to_units=pyunits.USD_2018))
            ),
        )
    print("\n----------Operation Cost----------")
    total_operating_cost = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.USD_2018 / pyunits.year
        )
    )
    print(f"Total Operating Cost: {total_operating_cost:.3f} $/year")

    opex_fraction = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.USD_2018 / pyunits.year
        )
        / pyunits.convert(
            m.fs.AD.inlet.flow_vol[0], to_units=pyunits.m**3 / pyunits.year
        )
        / m.fs.costing.LCOW
    )
    print(f"Operating cost fraction: {opex_fraction:.3f} $ opex / $ LCOW")

    print("Operating Cost Breakdown")
    for f in m.fs.costing.used_flows:
        print(
            f.title(),
            " :    {price:0.3f} $/m3 feed".format(
                price=value(
                    pyunits.convert(
                        m.fs.costing.aggregate_flow_costs[f],
                        to_units=pyunits.USD_2018 / pyunits.year,
                    )
                    / pyunits.convert(
                        m.fs.AD.inlet.flow_vol[0],
                        to_units=pyunits.m**3 / pyunits.year,
                    )
                )
            ),
        )

    print("\n----------Energy----------")

    electricity_intensity = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_electricity / m.fs.AD.inlet.flow_vol[0],
            to_units=units.kWh / units.m**3,
        )
    )
    print(f"Electricity Intensity: {electricity_intensity:.4f} kWh/m3")

    print("\n----------Levelized Cost----------")
    LCOW = value(
        pyunits.convert(m.fs.costing.LCOW, to_units=pyunits.USD_2018 / pyunits.m**3)
    )
    print(f"Levelized Cost of Water: {LCOW:.3f} $/m^3")


if __name__ == "__main__":
    m, results = build_flowsheet()
    assert_optimal_termination(results)
    stream_table = create_stream_table_dataframe(
        {
            "AD inlet": m.fs.AD.inlet,
            "AD liquid outlet": m.fs.AD.liquid_outlet,
            "AD vapor outlet": m.fs.AD.vapor_outlet,
            "Translator outlet": m.fs.translator_adm1_asm2d.outlet,
            "ElectroNP treated": m.fs.electroNP.treated,
            "ElectroNP byproduct": m.fs.electroNP.byproduct,
        },
        time_point=0,
    )
    print(stream_table_dataframe_to_string(stream_table))
    display_costing(m)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
)
from idaes.core.util.misc import StrEnum
from idaes.core.util.exceptions import ConfigurationError

"""
Ref: W. McGivney, S. Kawamura, Cost estimating manual for water treatment facilities, John Wiley & Sons, 2008. http://onlinelibrary.wiley.com/book/10.1002/9780470260036.
"""


class DewateringType(StrEnum):
    filter_belt_press = "filter_belt_press"
    filter_plate_press = "filter_plate_press"
    centrifuge = "centrifuge"


def cost_dewatering(
    blk, dewatering_type=DewateringType.centrifuge, cost_electricity_flow=True
):

    if dewatering_type == DewateringType.centrifuge:
        cost_centrifuge(blk, dewatering_type, cost_electricity_flow)

    elif dewatering_type == DewateringType.filter_belt_press:
        cost_filter_belt_press(blk, dewatering_type, cost_electricity_flow)

    elif dewatering_type == DewateringType.filter_plate_press:
        cost_filter_plate_press(blk, dewatering_type, cost_electricity_flow)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for dewatering_type:"
            f" {dewatering_type}. Argument must be a member of the DewateringType Enum class."
        )


def build_centrifuge_cost_param_block(blk):
    # NOTE: costing data are from McGivney & Kawamura, 2008
    blk.capital_a_parameter = pyo.Var(
        initialize=328.03,
        doc="A parameter for capital cost",
        units=pyo.units.USD_2007 / (pyo.units.gallon / pyo.units.hour),
    )
    blk.capital_b_parameter = pyo.Var(
        initialize=751295,
        doc="B parameter for capital cost",
        units=pyo.units.USD_2007,
    )


def build_filter_belt_press_cost_param_block(blk):
    # NOTE: costing data are from McGivney & Kawamura, 2008
    blk.capital_a_parameter = pyo.Var(
        initialize=146.29,
        doc="A parameter for capital cost",
        units=pyo.units.USD_2007 / (pyo.units.gallon / pyo.units.hour),
    )
    blk.capital_b_parameter = pyo.Var(
        initialize=433972,
        doc="B parameter for capital cost",
        units=pyo.units.USD_2007,
    )


def build_filter_plate_press_cost_param_block(blk):
    # NOTE: costing data are from McGivney & Kawamura, 2008
    blk.capital_a_parameter = pyo.Var(
        initialize=102794,
        doc="A parameter for capital cost",
        units=pyo.units.USD_2007 / (pyo.units.gallon / pyo.units.hour),
    )
    blk.capital_b_parameter = pyo.Var(
        initialize=0.4216,
        doc="B parameter for capital cost",
        units=pyo.units.dimensionless,
    )


@register_costing_parameter_block(
    build_rule=build_centrifuge_cost_param_block,
    parameter_block_name="centrifuge",
)
def cost_centrifuge(
    blk, dewatering_type=DewateringType.centrifuge, cost_electricity_flow=True
):
    """
    Centrifuge costing method
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    cost_blk = blk.costing_package.centrifuge
    t0 = blk.flowsheet().time.first()
    x = flow_in = pyo.units.convert(
        blk.unit_model.inlet.flow_vol[t0], to_units=pyo.units.gallon / pyo.units.hr
    )
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            cost_blk.capital_a_parameter * x + cost_blk.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )


@register_costing_parameter_block(
    build_rule=build_filter_belt_press_cost_param_block,
    parameter_block_name="filter_belt_press",
)
def cost_filter_belt_press(
    blk, dewatering_type=DewateringType.filter_belt_press, cost_electricity_flow=True
):
    """
    Belt Press Filter costing method
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    cost_blk = blk.costing_package.filter_belt_press
    t0 = blk.flowsheet().time.first()
    x = flow_in = pyo.units.convert(
        blk.unit_model.inlet.flow_vol[t0], to_units=pyo.units.gallon / pyo.units.hr
    )

    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            cost_blk.capital_a_parameter * x + cost_blk.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )


@register_costing_parameter_block(
    build_rule=build_filter_plate_press_cost_param_block,
    parameter_block_name="filter_plate_press",
)
def cost_filter_plate_press(
    blk, dewatering_type=DewateringType.filter_plate_press, cost_electricity_flow=True
):
    """
    Plate Press Filter costing method
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")

    cost_blk = blk.costing_package.filter_plate_press
    t0 = blk.flowsheet().time.first()
    x_units = pyo.units.gallon / pyo.units.hr
    x = flow_in = pyo.units.convert(blk.unit_model.inlet.flow_vol[t0], to_units=x_units)

    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            cost_blk.capital_a_parameter
            * x_units
            * (x / x_units) ** cost_blk.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )

    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a granular activated carbon unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from idaes.core.util.math import smooth_min
from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("GACZO")
class GACZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a granular activated carbon unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "gac"

        build_sido(self)

        # Empty Bed Contact Time
        self.empty_bed_contact_time = Var(
            units=pyunits.hour, bounds=(0, None), doc="Empty bed contact time of unit"
        )

        self._fixed_perf_vars.append(self.empty_bed_contact_time)
        self._perf_var_dict["Empty Bed Contact Time"] = self.empty_bed_contact_time

        # Electricity Demand
        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )

        self.electricity_intensity_parameter = Var(
            units=pyunits.kW / pyunits.m**3,
            doc="Parameter for calculating electricity based on empty bed "
            "contact time",
        )

        self.energy_electric_flow_vol_inlet = Var(
            units=pyunits.kWh / pyunits.m**3,
            doc="Electricity intensity with respect to inlet flowrate of unit",
        )

        @self.Constraint(doc="Electricity intensity based on empty bed contact time.")
        def electricity_intensity_constraint(b):
            return (
                b.energy_electric_flow_vol_inlet
                == b.electricity_intensity_parameter * b.empty_bed_contact_time
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on " "feed flowrate.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == (
                b.energy_electric_flow_vol_inlet
                * pyunits.convert(
                    b.get_inlet_flow(t), to_units=pyunits.m**3 / pyunits.hour
                )
            )

        self._fixed_perf_vars.append(self.electricity_intensity_parameter)

        self._perf_var_dict["Electricity Demand"] = self.electricity
        self._perf_var_dict[
            "Electricity Intensity"
        ] = self.energy_electric_flow_vol_inlet

        # Demand for activated carbon
        self.activated_carbon_replacement = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Replacement rate of activated carbon",
        )

        self.activated_carbon_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hour,
            bounds=(0, None),
            doc="Demand for activated carbon",
        )

        # Demand for activated carbon
        self.activated_carbon_bulk_density = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Bulk density, total mass of GAC per total bed volume",
        )

        @self.Constraint(
            self.flowsheet().time, doc="Constraint for activated carbon consumption."
        )
        def activated_carbon_equation(b, t):
            return b.activated_carbon_demand[t] == (
                b.activated_carbon_replacement
                * pyunits.convert(
                    b.get_inlet_flow(t), to_units=pyunits.m**3 / pyunits.hour
                )
            )

        self._fixed_perf_vars.append(self.activated_carbon_replacement)
        self._fixed_perf_vars.append(self.activated_carbon_bulk_density)

        self._perf_var_dict["Activated Carbon Demand"] = self.activated_carbon_demand

    @property
    def default_costing_method(self):
        return self.cost_gac

    @staticmethod
    def cost_gac(blk, number_of_parallel_units=5):
        """
        Adapted from core GAC costing model initially released in v0.6.0

        3 equation capital cost estimation for GAC systems with: (i), contactor/pressure vessel cost by polynomial
        as a function of individual contactor volume; (ii), initial charge of GAC adsorbent cost by exponential as a
        function of required mass of GAC adsorbent; and (iii), other process costs (vessels, pipes, instrumentation, and
        controls) calculated by power law as a function of total contactor(s) volume. Operating costs calculated as the
        required makeup and regeneration of GAC adsorbent. Energy for backwash and booster pumps considered negligible
        compared to regeneration costs
        Args:
            number_of_parallel_units (int, optional) - cost this unit as
                        number_of_parallel_units parallel units in operation (default: 5)
        """
        t0 = blk.flowsheet().time.first()

        Q = blk.unit_model.properties_in[t0].flow_vol
        T = blk.unit_model.empty_bed_contact_time
        gac_dens = blk.unit_model.activated_carbon_bulk_density

        # total bed volume
        V = pyo.units.convert(
            Q, to_units=pyo.units.m**3 / pyo.units.seconds
        ) * pyo.units.convert(T, to_units=pyo.units.seconds)

        # mass of gac in bed
        bed_mass_gac = pyo.units.convert(
            V, to_units=pyo.units.m**3
        ) * pyo.units.convert(gac_dens, to_units=pyo.units.kg / pyo.units.m**3)

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Call contactor, adsorbent (GAC) initial charge cost, and other process cost parameters
        (
            A0,
            A1,
            A2,
            A3,
            B0,
            B1,
            C0,
            C1,
            bed_mass_max_ref,
        ) = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "contactor_cost_coeff_0",
                "contactor_cost_coeff_1",
                "contactor_cost_coeff_2",
                "contactor_cost_coeff_3",
                "adsorbent_unit_cost_coeff",
                "adsorbent_unit_cost_exp_coeff",
                "other_cost_coeff",
                "other_cost_exp",
                "bed_mass_max_ref",
            ],
        )

        contactor_cost = number_of_parallel_units * pyo.units.convert(
            (
                A3 * (V / number_of_parallel_units) ** 3
                + A2 * (V / number_of_parallel_units) ** 2
                + A1 * (V / number_of_parallel_units) ** 1
                + A0
            ),
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        bed_mass_gac_ref = (
            smooth_min(
                bed_mass_max_ref / pyo.units.kg,
                pyo.units.convert(bed_mass_gac, to_units=pyo.units.kg) / pyo.units.kg,
            )
            * pyo.units.kg
        )

        adsorbent_unit_cost = pyo.units.convert(
            B0 * pyo.exp(bed_mass_gac_ref * B1),
            to_units=blk.config.flowsheet_costing_block.base_currency
            * pyo.units.kg**-1,
        )

        adsorbent_cost = adsorbent_unit_cost * bed_mass_gac

        other_process_cost = pyo.units.convert(
            (C0 * ((pyo.units.m**3) ** -C1) * V**C1),
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = contactor_cost + adsorbent_cost + other_process_cost

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        # electricity was not implemented in core GAC but retained in ZO update
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )

        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.activated_carbon_demand[t0], "activated_carbon"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a membrane aerated biofilm reactor unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("MABRZO")
class MABRZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a MABR unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "mabr"

        build_sido_reactive(self)

        self.nitrogen_removal_rate = Var(
            units=pyunits.g / pyunits.m**2 / pyunits.day,
            bounds=(0, None),
            doc="Nitrogen removal rate per day",
        )

        self._fixed_perf_vars.append(self.nitrogen_removal_rate)

        self.reactor_area = Var(
            units=pyunits.m**2,
            bounds=(0, None),
            doc="Sizing variable for effective reactor area",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for effective reactor area",
        )
        def reactor_area_constraint(b, t):
            return b.reactor_area == pyunits.convert(
                b.properties_treated[t].flow_mass_comp["ammonium_as_nitrogen"]
                / b.nitrogen_removal_rate,
                to_units=pyunits.m**2,
            )

        self._perf_var_dict["Reactor Area"] = self.reactor_area

        self.air_flow_rate = Var(
            self.flowsheet().config.time,
            units=pyunits.m**3 / pyunits.hour / pyunits.m**2,
            bounds=(0, None),
            doc="Air flow rate per area",
        )

        self._fixed_perf_vars.append(self.air_flow_rate)

        self.air_flow_vol = Var(
            self.flowsheet().config.time,
            units=pyunits.m**3 / pyunits.hour,
            bounds=(0, None),
            doc="Volumetric air flow rate",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for air flow",
        )
        def air_flow_constraint(b, t):
            return b.air_flow_vol[t] == pyunits.convert(
                b.air_flow_rate[t] * b.reactor_area,
                to_units=pyunits.m**3 / pyunits.hour,
            )

        self._perf_var_dict["Volumetric Air Flow Rate"] = self.air_flow_vol

        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )

        self._perf_var_dict["Electricity Demand"] = self.electricity

        self.energy_electric_flow_vol_inlet = Var(
            units=pyunits.kWh / pyunits.m**3,
            doc="Electricity intensity with respect to inlet flowrate of unit",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on air flowrate.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == (
                b.energy_electric_flow_vol_inlet
                * pyunits.convert(
                    b.air_flow_vol[t], to_units=pyunits.m**3 / pyunits.hour
                )
            )

        self._fixed_perf_vars.append(self.energy_electric_flow_vol_inlet)
        self._perf_var_dict[
            "Electricity Intensity"
        ] = self.energy_electric_flow_vol_inlet

    @property
    def default_costing_method(self):
        return self.cost_mabr

    @staticmethod
    def cost_mabr(blk):
        """
        General method for costing membrane aerated biofilm reactor. Capital cost
        is based on the cost of reactor and blower.
        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["reactor_cost", "blower_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        DCC_reactor = pyo.units.convert(
            blk.unit_model.properties_treated[t0].flow_mass_comp["ammonium_as_nitrogen"]
            / blk.unit_model.nitrogen_removal_rate
            * A,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        DCC_blower = pyo.units.convert(
            blk.unit_model.reactor_area * blk.unit_model.air_flow_rate[t0] * B,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        expr = DCC_reactor + DCC_blower

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    Expression,
    TransformationFactory,
)

from pyomo.network import Arc

from idaes.core import FlowsheetBlock
from idaes.core.util.scaling import calculate_scaling_factors
from idaes.core.util.initialization import propagate_state

from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    pretreatment_softening,
    costing,
)

from watertap.examples.flowsheets.full_treatment_train.model_components import (
    property_models,
)

from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)


def build_components(m):
    # build flowsheet
    pretrt_port = pretreatment_softening.build(m)
    property_models.build_prop(m, base="TDS")
    pretreatment_softening.build_tb(m)

    # Arc to translator block
    m.fs.s_pretrt_tb = Arc(
        source=pretrt_port["out"], destination=m.fs.tb_pretrt_to_desal.inlet
    )


def build(m):
    """
    Build a flowsheet with lime softening as the pretreatment process.
    """
    build_components(m)

    m.fs.treated_flow_vol = Expression(
        expr=m.fs.tb_pretrt_to_desal.properties_out[0].flow_vol
    )
    costing.build_costing(m)

    m.fs.removal_Ca = Expression(
        expr=(
            m.fs.stoich_softening_mixer_unit.inlet_stream_state[0.0].flow_mol_comp[
                "Ca(HCO3)2"
            ]
            - m.fs.stoich_softening_separator_unit.outlet_stream_state[
                0.0
            ].flow_mol_comp["Ca(HCO3)2"]
        )
        / m.fs.stoich_softening_mixer_unit.inlet_stream_state[0.0].flow_mol_comp[
            "Ca(HCO3)2"
        ]
    )
    m.fs.removal_Mg = Expression(
        expr=(
            m.fs.stoich_softening_mixer_unit.inlet_stream_state[0.0].flow_mol_comp[
                "Mg(HCO3)2"
            ]
            - m.fs.stoich_softening_separator_unit.outlet_stream_state[
                0.0
            ].flow_mol_comp["Mg(HCO3)2"]
        )
        / m.fs.stoich_softening_mixer_unit.inlet_stream_state[0.0].flow_mol_comp[
            "Mg(HCO3)2"
        ]
    )

    return m


def scale(m):
    pretreatment_softening.scale(m)
    calculate_scaling_factors(m.fs.tb_pretrt_to_desal)


def initialize(m):
    pretreatment_softening.initialize(m)
    propagate_state(m.fs.s_pretrt_tb)
    m.fs.tb_pretrt_to_desal.initialize()


def report(m):
    pretreatment_softening.display(m)
    m.fs.tb_pretrt_to_desal.report()


def solve_flowsheet():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    build(m)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    scale(m)
    calculate_scaling_factors(m)

    # initialize
    initialize(m)

    check_dof(m)
    solve_block(m, tee=True, fail_flag=True)

    # report
    report(m)

    return m


def simulate(m, check_termination=True):
    return solve_block(m, tee=False, fail_flag=check_termination)


if __name__ == "__main__":
    m = solve_flowsheet()
